/*
***********************************************************************
*                                                                     *          
*     File name : 64-bit.cc                                           *
*     Author    : Shuming Tan                                         *
*     Purpose   : rules checking places need to pay attention when    *
*                 porting code to 64-bit platforms.                   *
*                                                                     *
*     Copyright (C) 1996 by Abraxas Software. All rights reserved     *
*                                                                     *
***********************************************************************
*/


#include <check.cch>

/*
--------------------------------------------------------------------
  Use option -VP64 to specify if you are checking code to be reside
  on 64-bit bit machine. Otherwise, the rule assumes that the code
  is to reside on 32-bit machine.
--------------------------------------------------------------------
*/

int platform;

if (prj_begin)
{
    if (option('V'))
    {
        if (!strcmp(str_option('V'),"P64"))
        {
            platform=64;
        }
        else
        {
            platform=32;
        }
    }
    else
    {
        platform=32;
    }
}

////////////////////////////////////////////////////////////////////
//
//    Rule  1 : Always use int8, int4, int2, int1 where integers of
//              8-byte, 4-byte, 2-byte and 1-byte space allocation
//              is respectively implied.
//

/*
--------------------------------------------------------------------
  Assume that the naming convention for integer objects is like this

  the objects requiring size of 1-byte, 2-byte, 3-byte and 4-byte
  must start with i1, i2, i4 and i8 in their names correspondingly,
  or u1, u2, u4 and u4 if they they are also unsigned.
--------------------------------------------------------------------
*/

if (dcl_variable)
{
    if (strstr(dcl_name(),"i1")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"int1"))
        {
            warn(1001, "Use type int1 on 1-byte integer %s.",dcl_name());
        }
    }

    if (strstr(dcl_name(),"u1")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"uint1"))
        {
            warn(1001, "Use type uint1 on 1-byte unsigned integer %s.",dcl_name());
        }
    }
    if (strstr(dcl_name(),"i2")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"int2"))
        {
            warn(1001, "Use type int2 on 2-byte integer %s.",dcl_name());
        }
    }
    if (strstr(dcl_name(),"u2")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"uint2"))
        {
            warn(1001, "Use type uint2 on 2-byte unsigned integer %s.",dcl_name());
        }
    }
    if (strstr(dcl_name(),"i4")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"int4"))
        {
            warn(1001, "Use type int4 on 4-byte integer %s.",dcl_name());
        }
    }
    if (strstr(dcl_name(),"u4")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"uint4"))
        {
            warn(1001, "Use type uint4 on 4-byte unsigned integer %s.",dcl_name());
        }
    }
    if (strstr(dcl_name(),"i8")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"int8"))
        {
            warn(1001, "Use type int8 on 8-byte integer %s.",dcl_name());
        }
    }
    if (strstr(dcl_name(),"u8")==dcl_name())
    {
        if (strcmp(dcl_base_name(),"uint8"))
        {
            warn(1001, "Use type uint4 on 8-byte unsigned integer %s.",dcl_name());
        }
    }
}

////////////////////////////////////////////////////////////////////
//
//    Rule  2 : Do not use types directly; instead use typedefs to
//              user specific names.
//

/*
--------------------------------------------------------------------
  All types should not be used directly.

  Here assume that the 64-bit machine has types with size like this
  and the user specified type names are

  type           size      user defined type name

  char           1 bytes   int1
  unsigned char  1 bytes   uint1
  short          2 bytes   int2
  unsigned shot  2 bytes   uint2
  int            4 bytes   int4
  unsigned int   4 bytes   uint4
  long           8 bytes   int8
  unsigned long  8 bytes   uint8

  the 32-bit machine :

  type           size      user defined type name

  char           1 bytes   int1
  unsigned char  1 bytes   uint1
  short          2 bytes   int2
  unsigned shot  2 bytes   uint2
  int            4 bytes   int4
  unsigned int   4 bytes   uint4
  long           4 bytes   int4
  unsigned long  4 bytes   uint4

  these are subject to change depending on your need.

--------------------------------------------------------------------
*/  
if (dcl_variable||dcl_function/*||dcl_parameter*/)
{
    warn(9999,"name = %s",dcl_name());
    if (dcl_base==LONG_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type int8 instead of long.");
        }
        else
        {
            warn(1002, "Use type int4 instead of long.");
        }
    }
    if (dcl_base==ULONG_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type uint8 instead of unsigned long.");
        }
        else
        {
            warn(1002, "Use type uint4 instead of unsigned long.");
        }
    }
    if (dcl_base==INT_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type int4 instead of int.");
        }
        else
        {
            warn(1002, "Use type int4 instead of int.");
        }
    }
    if (dcl_base==SHORT_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type int2 instead of short.");
        }
        else
        {
            warn(1002, "Use type int2 instead of short.");
        }
    }
    if (dcl_base==CHAR_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type int1 instead of char.");
        }
        else
        {
            warn(1002, "Use type int1 instead of char.");
        }
    }
    if (dcl_base==UINT_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type uint4 instead of unsigned int.");
        }
        else
        {
            warn(1002, "Use type uint4 instead of unsigned int.");
        }
    }
    if (dcl_base==USHORT_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type uint2 instead of unsigned short.");
        }
        else
        {
            warn(1002, "Use type uint2 instead of unsigned short.");
        }
    }
    if (dcl_base==UCHAR_TYPE)
    {
        if (platform==64)
        {
            warn(1002, "Use type uint1 instead of unsigned char.");
        }
        else
        {
            warn(1002, "Use type uint1 instead of unsigned char.");
        }
    }
}

////////////////////////////////////////////////////////////////////
//
//    Rule  3 : Strictly eliminate use of unaligned accesses.
//

/*
--------------------------------------------------------------------
  This rule applies on the case where a object of type char or short
  is casted into an int or long type object.
--------------------------------------------------------------------
*/

if (op_cast)
{
    if (!strcmp(op_base_name(2),"int")
       ||!strcmp(op_base_name(2),"unsigned int")
       ||!strcmp(op_base_name(2),"long")
       ||!strcmp(op_base_name(2),"unsigned long"))
    {
        if (!strcmp(op_base_name(1),"char")
           ||!strcmp(op_base_name(1),"unsigned char")
           ||!strcmp(op_base_name(1),"short")
           ||!strcmp(op_base_name(1),"unsigned short"))
        {
            warn(1003, "Do cast a char or short object to int or long.");
        }
    }
}

////////////////////////////////////////////////////////////////////
//
//    Rule  4 : Do not assume the pointer and integer have the same 
//              size.
//

////////////////////////////////////////////////////////////////////
//
//    Rule  5 : Do not assume int and long have the same size.
//


////////////////////////////////////////////////////////////////////
//
//    Rule  6 : Do not pass long arguments to a function that expects
//              int arguments.
//

////////////////////////////////////////////////////////////////////
//
//    Rule  7 : Do not pass a pointer to a function that expect an
//              int as an argument.
//

int in_argument_list;

if (idn_function)
{
    in_argument_list=1;
}

if (op_call)
{
    in_argument_list=0;
}

if (cnv_any_to_ptr)
{
   if (strcmp(op_base_name(1),"int")==0)
   {
       warn(1004, "Do not assume the pointer and integer have the same size.");
   }
}

if (cnv_ptr_to_any)
{
   if (!in_argument_list)
   {
       if (strcmp(op_base_name(2),"int")==0)
       {
           warn(1004,"Do not assume the pointer and integer have the same size.");
       }
   }
   else
   {
       if (strcmp(op_base_name(2),"int")==0)
       {
           warn(1007, "Do not pass pointer to function that expects in argument.");
       }
   }
}


if (cnv_truncate)
{
    if (in_argument_list)
    {
        if (strcmp(op_base_name(2),"int")==0)
        {
            warn(1006,"Do not pass long argument to function that expects int argument.");
        }
    }
    else
    {
        if (strcmp(op_base_name(2),"int")==0)
        {
            warn(1005,"Do not assume int and long have the same size.");
        }
    }
}

////////////////////////////////////////////////////////////////////
//
//    Rule  8 : Use sizeof instead assuming the size of the types.
//

/*
--------------------------------------------------------------------
  Perhaps this rule is most often applied when calling function
  alloc or malloc to allocate a piece of memory of certain type.
  If operator sizeof is not found in actual argument list to the
  function call. It is assumed that the size of the type was assumed.
--------------------------------------------------------------------
*/
int in_alloc_call;
int sizeof_found;

if (idn_function)
{
    if (!strcmp(idn_name(),"malloc")||!strcmp(idn_name(),"alloc")||!strcmp(idn_name(),"realloc"))
    {
        in_alloc_call=1;
    }
}

if (op_sizeof)
{
    if (in_alloc_call)
    {
        sizeof_found=1;
    }
}

if (op_call)
{
    if (in_alloc_call)
    {
        if (!sizeof_found)
        {
            warn(1008, "Use sizeof instead assuming the size of the type.");
        }
    }
    sizeof_found=0;
    in_alloc_call=0;
}

////////////////////////////////////////////////////////////////////
//
//    Rule  9 : Always give the type specifiers for functions and
//              variables.
//              

if (dcl_no_specifier)
{
    warn(1009,"Declaration for %s need explicit type specifier.",dcl_name());
}

////////////////////////////////////////////////////////////////////
//
//    Rule 10 : Always use a long object to record the difference
//              between two pointers.
//

/*
--------------------------------------------------------------------
  This rule still trivial. It is only suitable for case such as
  diff=ptr1-ptr2;
--------------------------------------------------------------------
*/
int see_pointer_diff;

if (op_subt)
{
    if (op_level(1,0)==POINTER&&op_level(2,0)==POINTER)
    {
        see_pointer_diff=1;
    }
}

if (op_assign)
{
    if (see_pointer_diff)
    {
        if (strcmp(op_base_name(2),"long")
           ||strcmp(op_base_name(2),"unsigned long"))
        {
            warn(1010,"Always use long object to record difference btw pointers.");
        }
        see_pointer_diff=0;
    }
}

if (stm_end)
{
    see_pointer_diff=0;
}

////////////////////////////////////////////////////////////////////
//
//    Rule 11 : Always use %ld to format a long value in scanf and
//              printf, use %lu an unsigned long value.             
//

char *p;
char format[64];
int printf_or_scanf_call;
int format_number;
int format_string;
int arg_index;
int comma_number;

if (idn_function)
{
    if (idn_name())
    {
    if (!strcmp(idn_name(),"printf")||!strcmp(idn_name(),"scanf"))
    {
        printf_or_scanf_call=1;
    }
    if (!strcmp(idn_name(),"fprintf")||!strcmp(idn_name(),"fscanf"))
    {
        printf_or_scanf_call=2;
    }
    if (!strcmp(idn_name(),"sprintf")||!strcmp(idn_name(),"sscanf"))
    {
        printf_or_scanf_call=2;
    }
    comma_number=0;
    }
}

if (op_separator)
{
    if (printf_or_scanf_call)
    {
        if (comma_number==printf_or_scanf_call-1)
        {
            format_string=1;
        }
        comma_number++;
    }
}

if (op_call)
{
    if (printf_or_scanf_call)
    {
        if (comma_number==printf_or_scanf_call-1)
        {
            format_string=1;
        }
        arg_index=printf_or_scanf_call;
        while (arg_index<op_operands)
        {
            if (!strcmp(op_base_name(op_operands-arg_index),"long"))
            {
                if (format[format_number-(arg_index-printf_or_scanf_call)-1]!=1)
                {
                    warn(1011, "Use %%ld to format a long value for argument %d of function %s.", arg_index+1, op_function());
                }
            }
            if (!strcmp(op_base_name(op_operands-arg_index),"unsigned long"))
            {
                if (format[format_number-(arg_index-printf_or_scanf_call)-1]!=2)
                {
                    warn(1011, "Use %%lu to format a long value for argument %d of function %s.", arg_index+1, op_function());
                }
            }
            arg_index++;
        }
        printf_or_scanf_call=0;
    }
}

if (format_string)
{
    p=prev_token();
    format_number=0;
    while (strstr(p,"%"))
    {
        p=strstr(p,"%")+1;
        if (p[0]==0)
        {
            warn(1011, "Questionable format string ending with single %%.");
            break;
        }
        if (p[0]=='%')
        {
            p=p+1;
            continue;
        }
        format[format_number]=0;
        if (p[0]=='l')
        {
            if (p[1]=='d')
            {
                format[format_number]=1;
            }
            if (p[1]=='u')
            {
                format[format_number]=2;
            }
        }
        format_number++;
    }
} 
