/*
 *	Copyright (c) 1988-93 by Abraxas Software.
 *
 *	These rules compute the Halstead size metrics, for
 * 	both functions and modules.
 *
 *  Reference for formulas: IEEE Transactions on Software
 *  Engineering, vol. 14, No. 9, Sept. 1988.
 *
 *	REVISED 4-21-94 by Loren Cobb.
 *
 *  NOTE: The definition of Halstead operators and operands used
 *        by CodeCheck was corrected 9-24-93 for version 4.08.
 *        This will result in different (but more accurate)
 *        Halstead statistics reported by this rule file.
 */

statistic float   abstraction;    //  Halstead's abstraction level

float   total_tokens,     //  Total source-file tokens 
        unique_tokens,    //  Distinct source-file tokens
        fcn_length,       //  Halstead length measure (function)
        mod_length,       //  Halstead length measure (module)
		prj_length,       //  Halstead length measure (project)
		fcn_volume,       //  Halstead volume measure (function)
        mod_volume,       //  Halstead volume measure (module)
        prj_volume,       //  Halstead volume measure (project)
		potential_volume, //  Halstead potential volume (exact value - NOT approximate)
        effort,           //  Halstead effort (approximation)
		purity;           //  Halstead purity (predicted length / actual length)


if ( mod_begin )
	{
	printf( "\n\n------------------ %s -------------------\n\n", mod_name() );
	printf( "FUNCTION            Length    Volume  Abstraction\n" );
	reset( abstraction );
	}

if ( dcl_function )
	if ( dcl_definition )    //  Calculate the potential volume for this function:
		{
		if ( dcl_parm_count > 1 )
			{
			unique_tokens = dcl_parm_count + 3;
			total_tokens  = 2*dcl_parm_count + 1;
			}
		else
			{
			unique_tokens = dcl_parm_count + 2;
			total_tokens  = dcl_parm_count + 2;
			}
		potential_volume = total_tokens * log2( unique_tokens );
		}
	
if ( fcn_end )
	{
	if ( fcn_uH_operators == 0 )
		fcn_length = 0;
	else
		fcn_length = fcn_uH_operators * log2(fcn_uH_operators);
	if ( fcn_uH_operands > 0 )
		fcn_length += fcn_uH_operands * log2(fcn_uH_operands);

	total_tokens = fcn_H_operators + fcn_H_operands;
	unique_tokens = fcn_uH_operators + fcn_uH_operands;

	if ( unique_tokens > 1.0 )
		{
		fcn_volume = total_tokens * log2( unique_tokens );
		abstraction = 100 * potential_volume / fcn_volume;
		}
	else
		{
		fcn_volume = 0.0;
		abstraction = 0.0;
		}

	printf( "%-20s", fcn_name() );
	printf( "%6.0f", fcn_length );
	printf( "%10.0f", fcn_volume );
	printf( "%11.1f %%\n", abstraction );
	}

if ( mod_end )
	{
	total_tokens = mod_H_operators + mod_H_operands;
	unique_tokens = mod_uH_operators + mod_uH_operands;

	mod_volume =  total_tokens * log2( unique_tokens );	
	
	mod_length = mod_uH_operators * log2(mod_uH_operators) 
				+ mod_uH_operands * log2(mod_uH_operands);

	purity = mod_length / total_tokens;
	effort = (mod_uH_operators * mod_H_operands * mod_volume) / (2.0 * mod_uH_operands);
	
	printf( "\nLevel of Function Abstraction (percent)\nHistogram for module %s:\n", mod_name() );
	histogram( abstraction, 0, 50, 10 );
	
	printf( "\n\nModule %s:\n", mod_name() );
	printf( "   Halstead Volume:  %10.0f\n", mod_volume );
	printf( "   Halstead Length:  %10.0f\n", mod_length );
	printf( "   Halstead Purity:  %10.2f\n", purity );
	printf( "   Halstead Effort:  %10.0f\n", effort );
	}

if ( prj_end )
	if ( prj_modules > 1 )
		{
		printf( "\n\n=========================================\n\n", mod_name() );
	
		total_tokens = prj_H_operators + prj_H_operands;
		unique_tokens = prj_uH_operators + prj_uH_operands;
	
		prj_volume =  total_tokens * log2( unique_tokens );	
	
		prj_length = prj_uH_operators * log2(prj_uH_operators) 
			+ prj_uH_operands * log2(prj_uH_operands);
	
		purity = prj_length/total_tokens;
		effort = (prj_uH_operators * prj_H_operands * prj_volume) / (2.0 * prj_uH_operands);
		
		printf( "\nProject %s:\n", prj_name() );
		printf( "   Halstead Volume:  %15.0f\n", prj_volume );
		printf( "   Halstead Length:  %15.0f\n", prj_length );
		printf( "   Halstead Purity:  %15.2f\n", purity );
		printf( "   Halstead Effort:  %15.0f\n", effort );
		}
