/*	posix.cc
	Copyright (c) 1992-94 by Abraxas Software. All rights reserved.
==============================================================================
	Purpose:	Checks for POSIX.1 namespace infringement.
	Author:		Loren Cobb.
	Revision:	6 April 1992.
	Format:		Monospaced font with 4 spaces/tab.
==============================================================================

Abstract:

	These CodeCheck rules verify that the POSIX.1 reserved prefixes and
	suffixes have not been infringed by an application. These rules will
	also flag all redefinitions of POSIX identifiers, but only if the macro
	_POSIX_SOURCE is defined prior to the first #include.
	
	These rules do *NOT* check that every included system header contains
	exactly what is required by the POSIX standard. These rules are for
	checking application conformance, not implementation conformance.
	
	These rules are applied to all headers that are included in double
	quotes, e.g. #include "project.h". For proper use of these rules, be
	sure to include system headers in angle brackets: #include <ctypes.h>.

Warning Codes:

1000	Precede all headers with #define _POSIX_SOURCE.
1001	Leading underscores are reserved.
1002	Prefix <s> was used <n> times before this #include.
1003	Prefix <s> is reserved by <header>.
1004	Precede this declaration with #undef <name>
1005	This definition of <name> hides a prior definition.

Suggested Actions:

1000	The macro _POSIX_SOURCE must be defined for POSIX headers to be
		read correctly. Define this macro at the top of every source file.

1001	POSIX forbids the use of identifiers that begin with an underscore.
		Change the name of this identifier.

1002	The stated prefix is reserved by a POSIX header file, but has been
		found prior to its inclusion. You should find each preceding
		definition of an identifier with this prefix and change its name.
		To find each occurrence, move this #include to the top of the file
		and run this rule file again.

1003	The stated prefix or suffix is reserved by the named POSIX header
		file. Change the name of this identifier.

1004	POSIX requires that the identifier used here must be preceded with an
		#undef, to eliminate the possibility of unintended macro expansions.

1005	You have substantively redefined a macro that is already defined,
		possibly in a POSIX header. This is not prohibited by POSIX, but
		it is very poor style. To avoid confusion, change the macro name.

==============================================================================
*/

int	ch,					//  next character after prefix.
	ctype_included,		//  1 if header ctype.h		has been #included.
	dirent_included,	//  1 if header dirent.h	has been #included.
	errno_included,		//  1 if header errno.h		has been #included.
	fcntl_included,		//  1 if header fcntl.h		has been #included.
	grp_included,		//  1 if header grp.h		has been #included.
	limits_included,	//  1 if header limits.h	has been #included.
	locale_included,	//  1 if header locale.h	has been #included.
	pwd_included,		//  1 if header pwd.h		has been #included.
	signal_included,	//  1 if header signal.h	has been #included.
	string_included,	//  1 if header string.h	has been #included.
	sys_stat_included,	//  1 if header sys/stat.h	has been #included.
	sys_times_included,	//  1 if header sys/times.h	has been #included.
	termios_included,	//  1 if header termios.h	has been #included.
	prefix_is,			//  counts use of prefix  is
	prefix_to,			//  counts use of prefix  to
	prefix_d_,			//  counts use of prefix  d_
	prefix_E,			//	counts use of prefix  E     (followed by digit or uc letter)
	prefix_l_,			//  counts use of prefix  l_
	prefix_gr_,			//  counts use of prefix  gr_
	prefix_pw_,			//  counts use of prefix  pw_
	prefix_sa_,			//  counts use of prefix  sa_
	prefix_mem,			//  counts use of prefix  mem
	prefix_str,			//  counts use of prefix  str
	prefix_wcs,			//  counts use of prefix  wcs
	prefix_st_,			//  counts use of prefix  st_
	prefix_tms_,		//  counts use of prefix  tms_
	prefix_c_,			//  counts use of prefix  c_
	suffix_MAX,			//  counts use of suffix  _MAX
	posix_needed,		//  1 if _POSIX_SOURCE has not yet been defined.
	undef_needed;		//  1 if an #undef is needed in POSIX.


if ( mod_begin )
	{
	posix_needed = 1;
	set_option( 'S', 1 );	//	Make sure that project headers are checked, and
	}						//	that system headers are not checked (option -S1)

if ( pp_macro )
	{
	if ( prefix("_") )
		{
		if ( strcmp(pp_name(), "_POSIX_SOURCE") == 0 )
			posix_needed = 0;
		else
			warn( 1001, "%s: Leading underscores are reserved by POSIX.", dcl_name() );
		}
	}

if ( dcl_underscore )
	warn( 1001, "%s: Leading underscores are reserved by POSIX.", dcl_name() );

if ( header_name() )		// Triggers whenever an #include is processed.
	{
	if ( posix_needed )
		{
		warn( 1000, "Precede all headers with #define _POSIX_SOURCE" );
		posix_needed = 0;	//  This prevents multiple warnings.
		}
	if ( strcmp(header_name(),"ctype.h") == 0 )
		{
		if ( prefix_is )
			warn(1002, "Prefix \"is\"  was used %d times before this #include.", prefix_is );
		if ( prefix_to )
			warn(1002, "Prefix \"to\" was used %d times before this #include.", prefix_to );
		prefix_is = 0;
		prefix_to = 0;
		ctype_included = 1;
		}
	else if ( strcmp(header_name(),"dirent.h") == 0 )
		{
		if ( prefix_d_ )
			warn(1002, "Prefix d_ was used %d times before this #include.", prefix_d_ );
		prefix_d_ = 0;
		dirent_included = 1;
		}
	else if ( strcmp(header_name(),"errno.h") == 0 )
		{
		if ( prefix_E )
			warn(1002, "Prefix E was used %d times before this #include.", prefix_E );
		prefix_E = 0;
		errno_included = 1;
		}
	else if ( strcmp(header_name(),"fcntl.h") == 0 )
		{
		if ( prefix_l_ )
			warn(1002, "Prefix l_ was used %d times before this #include.", prefix_l_ );
		prefix_l_ = 0;
		fcntl_included = 1;
		}
	else if ( strcmp(header_name(),"grp.h") == 0 )
		{
		if ( prefix_gr_ )
			warn(1002, "Prefix gr_ was used %d times before this #include.", prefix_gr_ );
		prefix_gr_ = 0;
		grp_included = 1;
		}
	else if ( strcmp(header_name(),"limits.h") == 0 )
		{
		if ( suffix_MAX )
			warn(1002, "Suffix _MAX was used %d times before this #include.", suffix_MAX );
		suffix_MAX = 0;
		limits_included = 1;
		}
	else if ( strcmp(header_name(),"locale.h") == 0 )
		{
		locale_included = 1;
		}
	else if ( strcmp(header_name(),"pwd.h") == 0 )
		{
		if ( prefix_pw_ )
			warn(1002, "Prefix pw_ was used %d times before this #include.", prefix_pw_ );
		prefix_pw_ = 0;
		pwd_included = 1;
		}
	else if ( strcmp(header_name(),"signal.h") == 0 )
		{
		if ( prefix_sa_ )
			warn(1002, "Prefix sa_ was used %d times before this #include.", prefix_sa_ );
		prefix_sa_ = 0;
		signal_included = 1;
		}
	else if ( strcmp(header_name(),"string.h") == 0 )
		{
		if ( prefix_mem )
			warn(1002, "Prefix mem was used %d times before this #include.", prefix_mem );
		if ( prefix_str )
			warn(1002, "Prefix str was used %d times before this #include.", prefix_str );
		if ( prefix_wcs )
			warn(1002, "Prefix wcs was used %d times before this #include.", prefix_wcs );
		prefix_mem = 0;
		prefix_str = 0;
		prefix_wcs = 0;
		string_included = 1;
		}
	else if ( strcmp(header_name(),"sys/stat.h") == 0 )
		{
		if ( prefix_st_ )
			warn(1002, "Prefix st_ was used %d times before this #include.", prefix_st_ );
		prefix_st_ = 0;
		sys_stat_included = 1;
		}
	else if ( strcmp(header_name(),"sys/times.h") == 0 )
		{
		if ( prefix_tms_ )
			warn(1002, "Prefix tms_ was used %d times before this #include.", prefix_tms_ );
		prefix_tms_ = 0;
		sys_times_included = 1;
		}
	else if ( strcmp(header_name(),"termios.h") == 0 )
		{
		if ( prefix_c_ )
			warn(1002, "Prefix c_ was used %d times before this #include.", prefix_c_ );
		prefix_c_ = 0;
		termios_included = 1;
		}
	}

if ( dcl_ident_length || pp_macro )	//  Check each identifier for forbidden prefixes.
	{
	/*
		Check for reserved suffixes:
	*/
	if ( suffix("_t") )				//	DAL:519
		{
		warn( 1003, "%s: suffix _t is reserved by POSIX.", dcl_name() );
		}
	else if ( suffix("_MAX") )		//	DAL:527
		{
		++suffix_MAX;
		if ( limits_included )
			warn( 1003, "%s: suffix _MAX is reserved by limits.h", dcl_name() );
		}
	/*
		Check for reserved prefixes:
	*/
	if ( prefix("is") )				//	ANSI 4.13.2
		{
		++prefix_is;
		if ( ctype_included )
			warn( 1003, "%s: prefix \"is\" is reserved by ctype.h", dcl_name() );
		}
	else if ( prefix("to") )		//	ANSI 4.13.2
		{
		++prefix_to;
		if ( ctype_included )
			warn( 1003, "%s: prefix \"to\" is reserved by ctype.h", dcl_name() );
		}
	else if ( prefix("d_") )		//	DAL:522
		{
		++prefix_d_;
		if ( dirent_included )
			warn( 1003, "%s: prefix d_ is reserved by dirent.h", dcl_name() );
		}
	else if ( prefix("E") )			//	DAL:523
		{
		ch = root()[0];
		if ( isdigit(ch) || isupper(ch) )
			{
			++prefix_E;
			if ( errno_included )
				warn( 1003, "%s: prefix E is reserved by errno.h", dcl_name() );
			}
		}
	else if ( prefix("l_") )		//	DAL:524
		{
		++prefix_l_;
		if ( fcntl_included )
			warn( 1003, "%s: prefix l_ is reserved by fcntl.h", dcl_name() );
		}
	else if ( prefix("gr_") )		//	DAL:526
		{
		++prefix_gr_;
		if ( grp_included )
			warn( 1003, "%s: prefix gr_ is reserved by grp.h", dcl_name() );
		}
	else if ( prefix("pw_") )		//	DAL:521
		{
		++prefix_pw_;
		if ( pwd_included )
			warn( 1003, "%s: prefix pw_ is reserved by pwd.h", dcl_name() );
		}
	else if ( prefix("sa_") )		//	DAL:532
		{
		++prefix_sa_;
		if ( signal_included )
			warn( 1003, "%s: prefix sa_ is reserved by signal.h", dcl_name() );
		}
	else if ( prefix("mem") )		//	DAL:538
		{
		++prefix_mem;
		if ( string_included )
			warn( 1003, "%s: prefix mem is reserved by string.h", dcl_name() );
		}
	else if ( prefix("str") )		//	DAL:538
		{
		++prefix_str;
		if ( string_included )
			warn( 1003, "%s: prefix str is reserved by string.h", dcl_name() );
		}
	else if ( prefix("wcs") )		//	DAL:538
		{
		++prefix_wcs;
		if ( string_included )
			warn( 1003, "%s: prefix wcs is reserved by string.h", dcl_name() );
		}
	else if ( prefix("st_") )		//	DAL:539
		{
		++prefix_st_;
		if ( sys_stat_included )
			warn( 1003, "%s: prefix st_ is reserved by sys/stat.h", dcl_name() );
		}
	else if ( prefix("tms_") )		//	DAL:540
		{
		++prefix_tms_;
		if ( sys_times_included )
			warn( 1003, "%s: prefix tms_ is reserved by sys/times.h", dcl_name() );
		}
	else if ( prefix("c_") )		//	DAL:546
		{
		++prefix_c_;
		if ( termios_included )
			warn( 1003, "%s: prefix c_ is reserved by termios.h", dcl_name() );
		}
	/*
		Next, we check for prefixes that are
		allowed if preceded by an #undef.
	*/
	undef_needed = 0;
	if ( no_undef(dcl_name()) )
		{
		if ( prefix("F_") )				//	DAL:524
			{
			if ( fcntl_included )
				undef_needed = 1;
			}
		else if ( prefix("S_") )		//	DAL:524 & 539
			{
			if ( fcntl_included || sys_stat_included )
				undef_needed = 1;
			}
		else if ( prefix("O_") )		//	DAL:524
			{
			if ( fcntl_included )
				undef_needed = 1;
			}
		else if ( prefix("LC_") )		//	ANSI 4.13.3
			{
			ch = root()[0];
			if ( locale_included && (isdigit(ch) || isupper(ch)) )
				undef_needed = 1;
			}
		else if ( prefix("SIG") )		//	DAL:532
			{
			if ( signal_included )
				undef_needed = 1;
			}
		else if ( prefix("SA_") )		//	DAL:532
			{
			if ( signal_included )
				undef_needed = 1;
			}
		else if ( prefix("V") )			//	DAL:546
			{
			if ( termios_included )
				undef_needed = 1;
			}
		else if ( prefix("I") )			//	DAL:546
			{
			if ( termios_included )
				undef_needed = 1;
			}
		else if ( prefix("O") )			//	DAL:546
			{
			if ( termios_included )
				undef_needed = 1;
			}
		else if ( prefix("TC") )		//	DAL:546
			{
			if ( termios_included )
				undef_needed = 1;
			}
		else if ( prefix("B") )			//	DAL:546
			{
			ch = root()[0];
			if ( termios_included && isdigit(ch) )
				undef_needed = 1;
			}

		if ( undef_needed )
			warn( 1004, "Precede this line with #undef %s", dcl_name() );
		}
	}

if ( pp_stack )
	warn( 1005, "This definition of %s hides a prior definition.", pp_name() );
