/*	SVID.cc

	Copyright (c) 1992-94 by Abraxas Software. All rights reserved.
==============================================================================
	Purpose:	Flags SVID functions that have POSIX.1 replacements.
	Author:		Loren Cobb.
	Revision:	12 October 1994.
	Format:		Monospaced font with 4 spaces/tab.
==============================================================================

Abstract:

	These CodeCheck rules generate warning messages when SVID features
	are used that are not POSIX conforming. If possible, these messages
	will suggest the appropriate POSIX feature to use.
	
	The features flagged in these rules are the ones that I know about as
	of the date on this rule file. If you know of others that ought to be
	flagged, please fax me the details at your earliest convenience. The
	Abraxas fax number is 503-244-8375. Many thanks!

Warning Codes:

3000	Precede all headers with #define _POSIX_SOURCE.
3001	Replace function <SVID function> with <POSIX function>.
3002	Function <SVID function> has no POSIX equivalent.
3003	Function <SVID function> is not needed in POSIX.
3004	POSIX requires #include <POSIX header> for <function>.
3005	Replace <SVID header> with <POSIX header>.
3006	Replace <SVID macro> with <POSIX macro>.
3007	If you need an audible alarm, use \a instead of \07.


Suggested Actions:

3000	The macro _POSIX_SOURCE must be defined for POSIX headers to be
		read correctly. Define this macro at the top of every source file.
3001    Look up the replacement function in a POSIX reference - the return
		type and some arguments may differ from the SVID function, and a
		POSIX header may need to be included.
3002	You will need to hand-code a replacement function.
3003	No call to this function is needed in a standard POSIX environment.
3004	Insert the appropriate #include after #define _POSIX_SOURCE.
3005	Change the name of the SVID header to the appropriate POSIX header.
3006	Change the name of the SVID macro to the appropriate POSIX macro.
3007	If you need an audible alarm, change \07 to \a in the string.


Useful References:

Horton, Mark R. (1990) "Portable C Software." Published by Prentice-Hall,
		Englewood Cliffs, NJ 07632, USA.

IEEE (1988) "IEEE Standard Portable Operating System Interface for Computer
		Environments 1003.1-1988." Published by IEEE, 345 East 47th Street,
		New York, NY 10017, USA.

Lewine, Donald A. (1991) "POSIX Programmer's Guide."  Published by O'Reilly &
		Associates, 632 Petaluma Avenue, Sebastopol, CA 95472, USA.

Zlotnick, Fred (1991) "The POSIX.1 Standard." Published by Benjamin/Cummings,
		390 Bridge Parkway, Redwood City, CA 94065, USA.

==============================================================================
*/

#define		NEED_POSIX		3000
#define		REPLACE_FCN		3001
#define		NO_EQUIV		3002
#define		NOT_NEEDED		3003
#define		INCLUDE			3004
#define		REPLACE_HDR		3005
#define		REPLACE_CON		3006
#define		ALARM			3007

#define REPLACE(fname,gname)  if ( strcmp(op_function(),fname) == 0 )		\
			{																\
			warn(REPLACE_FCN, "Replace " fname " with " gname ".");			\
			++items_flagged;												\
			}

#define REWRITE(fname)  if ( strcmp(op_function(),fname) == 0 )				\
			{																\
			warn(NO_EQUIV, "Function " fname " has no POSIX equivalent.");	\
			++items_flagged;												\
			}

#define DELETE(fname)  if ( strcmp(op_function(),fname) == 0 )				\
			{																\
			warn(NOT_NEEDED, "Function " fname " is not needed in POSIX.");	\
			++items_flagged;												\
			}

#define CALLED(fname)  (strcmp(op_function(),fname) == 0)


int		posix_needed,		//  1 if macro _POSIX_SOURCE has not yet been defined.
		unistd_needed,		//  1 if header unistd.h	has not yet been included.
		items_flagged;		//  Number of non-POSIX features found in this module.

int		unistd_included;	//  1 if header unistd.h	has been #included.


if ( mod_begin )
	{
	posix_needed		= 1;
	unistd_needed		= 1;
	items_flagged		= 0;
	unistd_included		= 0;
	}

if ( pp_macro )
	if ( strcmp(pp_name(), "_POSIX_SOURCE") == 0 )
		posix_needed = 0;

if ( header_name() )
	{
	if ( posix_needed )
		{
		warn( NEED_POSIX, "Precede all headers with #define _POSIX_SOURCE" );
		posix_needed = 0;
		++items_flagged;
		}

	if ( unistd_needed )
		{
		if ( strcmp(header_name(), "unistd.h") != 0 )
			warn( INCLUDE, "POSIX recommends #include <unistd.h> before this line." );
		unistd_needed = 0;
		++items_flagged;
		}

	if ( strcmp(header_name(),"unistd.h") == 0 )
		{
		unistd_included = 1;
		unistd_needed = 0;
		}
	else if ( strcmp(header_name(),"varargs.h") == 0 )
		{
		warn( REPLACE_HDR, "Replace <varargs.h> with <stdargs.h>." );
		++items_flagged;
		}
	}

if ( op_call )
	{
		 REPLACE( "cuserid"			, "getlogin or getpwuid"	)		//  DAL:249
	else REPLACE( "drand48"			, "rand"					)		//  DAL:568
	else REPLACE( "erand48"			, "rand"					)		//  DAL:568
	else REPLACE( "ftw"				, "opendir, readdir, or closedir" )	//  DAL:568
	else REPLACE( "getw"			, "fread"					)		//  DAL:568
	else REPLACE( "ioctl"			, "[many possibilities]"	)		//  DAL:568
	else REPLACE( "jrand48"			, "rand"					)		//  DAL:568
	else REPLACE( "lockf"			, "fcntl"					)		//  DAL:568
	else REPLACE( "lrand48"			, "rand"					)		//  DAL:568
	else REPLACE( "mknod"			, "mkdir or mkfifo"			)		//  DAL:568
	else REPLACE( "popen"			, "pipe, fdopen, fork, system, or wait"	)	//  DAL:568
	else REPLACE( "signal"			, "sigaction"				)		//  DAL:420
	else REPLACE( "srand48"			, "srand"					)		//  DAL:568

	else REWRITE( "erf"				)		//  DAL:568
	else REWRITE( "erfc"			)		//  DAL:568
	else REWRITE( "gamma"			)		//  DAL:568
	else REWRITE( "getopt"			)		//  DAL:568
	else REWRITE( "hcreate"			)		//  DAL:568
	else REWRITE( "hdestroy"		)		//  DAL:568
	else REWRITE( "hsearch"			)		//  DAL:568
	else REWRITE( "hypot"			)		//  DAL:568
	else REWRITE( "isascii"			)		//  DAL:568
	else REWRITE( "j0"				)		//  DAL:568
	else REWRITE( "j1"				)		//  DAL:568
	else REWRITE( "jn"				)		//  DAL:568
	else REWRITE( "lfind"			)		//  DAL:568
	else REWRITE( "lsearch"			)		//  DAL:568
	else REWRITE( "remque"			)		//  DAL:568
	else REWRITE( "tdelete"			)		//  DAL:568
	else REWRITE( "tfind"			)		//  DAL:568
	else REWRITE( "tsearch"			)		//  DAL:568
	else REWRITE( "twalk"			)		//  DAL:568
	else REWRITE( "y0"				)		//  DAL:568
	else REWRITE( "y1"				)		//  DAL:568
	else REWRITE( "yn"				)		//  DAL:568
	}

if ( macro("O_NDELAY") )		//  MRH:328
	{
	warn( REPLACE_CON, "Replace O_NDELAY with O_NONBLOCK." );
	++items_flagged;
	}

if ( macro("SIGIOT") )		//  DAL:211
	{
	warn( REPLACE_CON, "Replace SIGIOT with SIGABRT." );
	++items_flagged;
	}

if ( lex_num_escape == 7 )	//  DAL:377
	{
	warn( ALARM, "If you need an audible alarm, use \\a instead of \\07." );
	++items_flagged;
	}

if ( mod_end )
	{
	printf( "\n*** %d non-POSIX SVID features were found in module %s ***\n\n", items_flagged, mod_name() );
	}
