/*	wrapper.cc

	Copyright (c) 1992-93 by Abraxas Software. All rights reserved.
==============================================================================
Purpose:	Detects header files that do not have wrappers, and
			header #include directives that are not conditional.
Author:		Loren Cobb.
Revision:	21 September 1992.
Format:		Monospaced font with 4 spaces/tab.
Thanks to:	Dan Sullivan, Amaryllis Software, for this idea.
==============================================================================

Introduction:

	Many C programmers like to place "wrappers" around the contents of
	their header files. The wrapper guarrantees that the contents of the
	header file will be read exactly once by the compiler. A wrapped header
	named foobar.h typically looks like this:
	
		#ifndef __FOOBAR__		//	First non-comment line of foobar.h
		#define __FOOBAR__
		
		//  All foobar contents appear here...
		
		#endif					//	Last line of foobar.h
	
	Once a header file has been wrapped, it is also possible to use the
	wrapper to make sure that the header is opened only once. This further
	improves the compilation speed of the program. Simply enclose the
	#include directive within a conditional test, like this:
	
		#ifndef __FOOBAR__
		#include "foobar.h"
		#endif


Abstract:

	These rules first verify that each header file has a wrapper that is
	constructed as follows: delete the filename extension, convert the
	name to uppercase, and then append two underscores onto the beginning
	and end of the name. If your convention for wrapper naming differs from
	this, you will have to rewrite part of the last rule in this file.
	
	Second, these rules verify that every #include that is found within
	a header file is enclosed within a conditional preprocessor directive.


System vs. User Headers:

	As supplied, these rules are applied to EVERY header that is included.
	If you want to restrict these rules to user-written header files, then
	do this:
	
	1.	Change the set_option call in the first rule to
	
		set_option( 'S', 1 );
	
	2.	Make sure that every user-written header file is included with
	double quotes, like this:
		
		#include "foobar.h"
		
	and NOT like this:
		
		#include <foobar.h>
		
	Note: It is a long-standing C convention that system headers are to
	be included with angle brackets, and user headers are to be included
	with double quotes.
	

Warning Codes:

1001	This #include should be wrapped in an #ifndef.
1002	This header file needs a wrapper.
1003	<name> is not the correct wrapper name for this file!
1004	This #define should be wrapped within an #ifndef.

==============================================================================
*/


int		lin_if_depth,		//  Holds latest value of pp_if_depth.
		wrapper_needed,		//  True when a header file is opened.
		bad_name,			//  True if the wrapper name is wrong.
		length, k, ch1, ch2;


if ( mod_begin )
	{
	set_option( 'S', 3 );	//  Apply these rules to ALL headers.
	}

if ( pp_if_depth || pp_endif )
	lin_if_depth = pp_if_depth;

if ( pp_include )
	{
	wrapper_needed = 1;
	if ( lin_header && lin_if_depth == 0 )
		warn( 1001, "This #include should be wrapped in an #ifndef." );
	}

if ( lin_has_code )
	if ( lin_header && wrapper_needed )
		{
		warn( 1002, "This header file needs a wrapper." );
		wrapper_needed = 0;
		}

if ( pp_macro )
	if ( lin_header && wrapper_needed )
		{
		wrapper_needed = 0;
		bad_name = 0;
		length = strlen(pp_name());
		if ( pp_name()[0] != '_' )
			bad_name = 1;
		else if ( pp_name()[1] != '_' )
			bad_name = 1;
		else if ( pp_name()[length-2] != '_' )
			bad_name = 1;
		else if ( pp_name()[length-1] != '_' )
			bad_name = 1;
		else
			{
			k = 2;
			while ( k < length-2 )
				{
				ch1 = pp_name()[k];
				ch2 = file_name()[k-2];
				if ( ch1 != toupper(ch2) )
					{
					bad_name = 1;
					break;
					}
				k++;
				}
			}
		if ( bad_name )
			warn( 1003, "%s is not the correct wrapper name for this file!", pp_name() );
		else if ( lin_if_depth == 0 )
			warn( 1004, "This #define should be wrapped within an #ifndef." );
		}
